"""Type stubs for claim169 Python bindings."""

from typing import Callable, Literal, Optional

class Claim169Exception(Exception):
    """Base exception for Claim 169 errors."""
    ...

class Base45DecodeError(Claim169Exception):
    """Base45 decoding failed."""
    ...

class DecompressError(Claim169Exception):
    """Zlib decompression failed."""
    ...

class CoseParseError(Claim169Exception):
    """COSE parsing failed."""
    ...

class CwtParseError(Claim169Exception):
    """CWT parsing failed."""
    ...

class Claim169NotFoundError(Claim169Exception):
    """Claim 169 not found in CWT."""
    ...

class SignatureError(Claim169Exception):
    """Signature verification failed."""
    ...

class DecryptionError(Claim169Exception):
    """Decryption failed."""
    ...

class EncryptionError(Claim169Exception):
    """Encryption failed."""
    ...

class Biometric:
    """Biometric data extracted from claim 169."""
    data: bytes
    format: Optional[int]
    sub_format: Optional[int]
    issuer: Optional[str]

class CwtMeta:
    """CWT metadata (issuer, subject, timestamps)."""
    issuer: Optional[str]
    subject: Optional[str]
    expires_at: Optional[int]
    not_before: Optional[int]
    issued_at: Optional[int]

    def is_valid_now(self) -> bool:
        """Check if token is currently valid."""
        ...

    def is_expired(self) -> bool:
        """Check if token is expired."""
        ...

class CertificateHash:
    """X.509 certificate hash (COSE_CertHash)."""
    algorithm: str
    """Hash algorithm identifier (numeric COSE algorithm ID as string or named algorithm)."""
    hash_value: bytes
    """Hash value bytes."""

class X509Headers:
    """X.509 headers extracted from COSE protected/unprotected headers."""
    x5bag: Optional[list[bytes]]
    """x5bag (label 32): Unordered bag of X.509 certificates (DER-encoded)."""
    x5chain: Optional[list[bytes]]
    """x5chain (label 33): Ordered chain of X.509 certificates (DER-encoded)."""
    x5t: Optional[CertificateHash]
    """x5t (label 34): Certificate thumbprint hash."""
    x5u: Optional[str]
    """x5u (label 35): URI pointing to an X.509 certificate."""

    def has_any(self) -> bool:
        """Check if any X.509 headers are present."""
        ...

class Claim169:
    """Decoded Claim 169 identity data."""
    id: Optional[str]
    version: Optional[str]
    language: Optional[str]
    full_name: Optional[str]
    first_name: Optional[str]
    middle_name: Optional[str]
    last_name: Optional[str]
    date_of_birth: Optional[str]
    gender: Optional[int]
    address: Optional[str]
    email: Optional[str]
    phone: Optional[str]
    nationality: Optional[str]
    marital_status: Optional[int]
    guardian: Optional[str]
    photo: Optional[bytes]
    photo_format: Optional[int]
    best_quality_fingers: Optional[bytes]
    secondary_full_name: Optional[str]
    secondary_language: Optional[str]
    location_code: Optional[str]
    legal_status: Optional[str]
    country_of_issuance: Optional[str]

    # Biometrics
    right_thumb: Optional[list[Biometric]]
    right_pointer_finger: Optional[list[Biometric]]
    right_middle_finger: Optional[list[Biometric]]
    right_ring_finger: Optional[list[Biometric]]
    right_little_finger: Optional[list[Biometric]]
    left_thumb: Optional[list[Biometric]]
    left_pointer_finger: Optional[list[Biometric]]
    left_middle_finger: Optional[list[Biometric]]
    left_ring_finger: Optional[list[Biometric]]
    left_little_finger: Optional[list[Biometric]]
    right_iris: Optional[list[Biometric]]
    left_iris: Optional[list[Biometric]]
    face: Optional[list[Biometric]]
    right_palm: Optional[list[Biometric]]
    left_palm: Optional[list[Biometric]]
    voice: Optional[list[Biometric]]

    def has_biometrics(self) -> bool:
        """Check if claim has biometric data."""
        ...

    def to_dict(self) -> dict:
        """Convert to dictionary."""
        ...

class DecodeResult:
    """Result of decoding a Claim 169 QR code."""
    claim169: Claim169
    cwt_meta: CwtMeta
    verification_status: str
    x509_headers: X509Headers
    """X.509 headers from COSE protected/unprotected headers."""

    def is_verified(self) -> bool:
        """Check if signature was verified."""
        ...

# ============================================================================
# Callback Types for Custom Crypto Providers
# ============================================================================

# Signature verifier callback type
# Use for external crypto providers (HSM, cloud KMS, remote signing, etc.)
VerifierCallback = Callable[
    [str, Optional[bytes], bytes, bytes],  # (algorithm, key_id, data, signature)
    None  # Returns None on success, raises exception on failure
]

# Decryptor callback type
# Use for external crypto providers (HSM, cloud KMS, etc.)
DecryptorCallback = Callable[
    [str, Optional[bytes], bytes, bytes, bytes],  # (algorithm, key_id, nonce, aad, ciphertext)
    bytes  # Returns decrypted plaintext
]

# Signer callback type
# Use for external crypto providers (HSM, cloud KMS, remote signing, etc.)
SignerCallback = Callable[
    [str, Optional[bytes], bytes],  # (algorithm, key_id, data)
    bytes  # Returns signature bytes
]

# Encryptor callback type
# Use for external crypto providers (HSM, cloud KMS, etc.)
EncryptorCallback = Callable[
    [str, Optional[bytes], bytes, bytes, bytes],  # (algorithm, key_id, nonce, aad, plaintext)
    bytes  # Returns ciphertext with auth tag
]

# Algorithm type hints
SignAlgorithm = Literal["EdDSA", "ES256"]
EncryptAlgorithm = Literal["A256GCM", "A128GCM"]

# ============================================================================
# Crypto Hook Wrapper Classes
# ============================================================================

class PySignatureVerifier:
    """Wrapper for custom signature verifier callback."""
    def __init__(self, callback: VerifierCallback) -> None: ...

class PyDecryptor:
    """Wrapper for custom decryptor callback."""
    def __init__(self, callback: DecryptorCallback) -> None: ...

class PySigner:
    """Wrapper for custom signer callback.

    Use this for external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Remote signing services
    - Smart cards and TPMs
    """
    def __init__(
        self,
        callback: SignerCallback,
        key_id: Optional[bytes] = None,
    ) -> None: ...

class PyEncryptor:
    """Wrapper for custom encryptor callback.

    Use this for external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Custom software keystores
    """
    def __init__(self, callback: EncryptorCallback) -> None: ...

# ============================================================================
# Decode Functions
# ============================================================================

def decode_unverified(
    qr_text: str,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
) -> DecodeResult:
    """
    Decode a Claim 169 QR code without signature verification (INSECURE).

    Args:
        qr_text: The QR code text content (Base45 encoded)
        skip_biometrics: If True, skip decoding biometric data
        max_decompressed_bytes: Maximum decompressed size
        validate_timestamps: If True, validate exp/nbf timestamps (default: True)
        clock_skew_tolerance_seconds: Tolerance for timestamp validation (default: 0)

    Returns:
        DecodeResult containing the decoded claim and CWT metadata

    Raises:
        Base45DecodeError: If Base45 decoding fails
        DecompressError: If zlib decompression fails
        CoseParseError: If COSE parsing fails
        CwtParseError: If CWT parsing fails
        Claim169NotFoundError: If claim 169 is not present
    """
    ...

def decode(
    qr_text: str,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
    verify_with_ed25519: Optional[bytes] = None,
    verify_with_ecdsa_p256: Optional[bytes] = None,
    verify_with_ed25519_pem: Optional[str] = None,
    verify_with_ecdsa_p256_pem: Optional[str] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode a Claim 169 QR code.

    Security:
    - By default, requires signature verification via one of:
      - `verify_with_ed25519` (32 bytes)
      - `verify_with_ecdsa_p256` (33 or 65 bytes SEC1)
      - `verify_with_ed25519_pem` (PEM string)
      - `verify_with_ecdsa_p256_pem` (PEM string)
    - To explicitly decode without verification (testing only), set
      `allow_unverified=True`.
    """
    ...

def decode_with_ed25519(
    qr_text: str,
    public_key: bytes,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
) -> DecodeResult:
    """
    Decode with Ed25519 signature verification.

    Args:
        qr_text: The QR code text content
        public_key: Ed25519 public key bytes (32 bytes)
    """
    ...

def decode_with_ecdsa_p256(
    qr_text: str,
    public_key: bytes,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
) -> DecodeResult:
    """
    Decode with ECDSA P-256 signature verification.

    Args:
        qr_text: The QR code text content
        public_key: SEC1 encoded P-256 public key bytes (33 or 65 bytes)
    """
    ...

def decode_with_ed25519_pem(
    qr_text: str,
    pem: str,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
) -> DecodeResult:
    """
    Decode with Ed25519 signature verification using PEM format.

    Args:
        qr_text: The QR code text content
        pem: PEM-encoded Ed25519 public key (SPKI format)
    """
    ...

def decode_with_ecdsa_p256_pem(
    qr_text: str,
    pem: str,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
) -> DecodeResult:
    """
    Decode with ECDSA P-256 signature verification using PEM format.

    Args:
        qr_text: The QR code text content
        pem: PEM-encoded ECDSA P-256 public key (SPKI format)
    """
    ...

def decode_with_verifier(qr_text: str, verifier: VerifierCallback) -> DecodeResult:
    """
    Decode with a custom verifier callback.

    Use for integration with external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Remote signing services
    - Smart cards and TPMs

    Args:
        qr_text: The QR code text content
        verifier: Callable that verifies signatures

    Example:
        def my_verify(algorithm, key_id, data, signature):
            crypto_provider.verify(key_id, data, signature)

        result = decode_with_verifier(qr_text, my_verify)
    """
    ...

def decode_encrypted_aes(
    qr_text: str,
    key: bytes,
    verifier: Optional[VerifierCallback] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode an encrypted Claim 169 QR code with AES-GCM.

    Supports both AES-128 and AES-256 based on key length.

    Args:
        qr_text: The QR code text content
        key: AES-GCM key bytes (16 bytes for AES-128, 32 bytes for AES-256)
        verifier: Optional verifier for nested signature verification
        allow_unverified: If True, allow decoding without signature verification (INSECURE)
    """
    ...

def decode_encrypted_aes256(
    qr_text: str,
    key: bytes,
    verifier: Optional[VerifierCallback] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode an encrypted Claim 169 QR code with AES-256-GCM.

    Args:
        qr_text: The QR code text content
        key: AES-256 key bytes (32 bytes)
        verifier: Optional verifier for nested signature verification
        allow_unverified: If True, allow decoding without signature verification (INSECURE)
    """
    ...

def decode_encrypted_aes128(
    qr_text: str,
    key: bytes,
    verifier: Optional[VerifierCallback] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode an encrypted Claim 169 QR code with AES-128-GCM.

    Args:
        qr_text: The QR code text content
        key: AES-128 key bytes (16 bytes)
        verifier: Optional verifier for nested signature verification
        allow_unverified: If True, allow decoding without signature verification (INSECURE)
    """
    ...

def decode_with_decryptor(
    qr_text: str,
    decryptor: DecryptorCallback,
    verifier: Optional[VerifierCallback] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode encrypted with a custom decryptor callback.

    Use for integration with external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Custom software keystores

    Args:
        qr_text: The QR code text content
        decryptor: Callable that decrypts ciphertext
        verifier: Optional verifier for nested signature verification
        allow_unverified: If True, allow decoding without signature verification (INSECURE)

    Example:
        def my_decrypt(algorithm, key_id, nonce, aad, ciphertext):
            return crypto_provider.decrypt(key_id, nonce, aad, ciphertext)

        result = decode_with_decryptor(qr_text, my_decrypt, allow_unverified=True)
    """
    ...

def version() -> str:
    """Get the library version."""
    ...

# ============================================================================
# Encoder Input Classes
# ============================================================================

class Claim169Input:
    """Input data for encoding a Claim 169 credential.

    Only `id` and `full_name` can be set in the constructor.
    Other fields must be set as attributes after construction.
    """

    # Settable attributes
    id: Optional[str]
    version: Optional[str]
    language: Optional[str]
    full_name: Optional[str]
    first_name: Optional[str]
    middle_name: Optional[str]
    last_name: Optional[str]
    date_of_birth: Optional[str]
    gender: Optional[int]
    address: Optional[str]
    email: Optional[str]
    phone: Optional[str]
    nationality: Optional[str]
    marital_status: Optional[int]
    guardian: Optional[str]
    photo: Optional[bytes]
    photo_format: Optional[int]
    secondary_full_name: Optional[str]
    secondary_language: Optional[str]
    location_code: Optional[str]
    legal_status: Optional[str]
    country_of_issuance: Optional[str]

    def __init__(
        self,
        id: Optional[str] = None,
        full_name: Optional[str] = None,
    ) -> None: ...

class CwtMetaInput:
    """Input metadata for encoding a CWT token.

    All fields can be set as attributes after construction.
    """

    # Settable attributes
    issuer: Optional[str]
    subject: Optional[str]
    expires_at: Optional[int]
    not_before: Optional[int]
    issued_at: Optional[int]

    def __init__(
        self,
        issuer: Optional[str] = None,
        expires_at: Optional[int] = None,
    ) -> None: ...

# ============================================================================
# Encoder Functions
# ============================================================================

def encode_with_ed25519(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    private_key: bytes,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with Ed25519 signature.

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        private_key: Ed25519 private key bytes (32 bytes)
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Raises:
        ValueError: If private key is invalid
        Claim169Exception: If encoding fails
    """
    ...

def encode_with_ecdsa_p256(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    private_key: bytes,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with ECDSA P-256 signature.

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        private_key: ECDSA P-256 private key bytes (32 bytes)
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Raises:
        ValueError: If private key is invalid
        Claim169Exception: If encoding fails
    """
    ...

def encode_signed_encrypted(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    sign_key: bytes,
    encrypt_key: bytes,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with Ed25519 signature and AES-256-GCM encryption.

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        sign_key: Ed25519 private key bytes (32 bytes)
        encrypt_key: AES-256 key bytes (32 bytes)
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Raises:
        ValueError: If keys are invalid
        Claim169Exception: If encoding fails
    """
    ...

def encode_signed_encrypted_aes128(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    sign_key: bytes,
    encrypt_key: bytes,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with Ed25519 signature and AES-128-GCM encryption.

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        sign_key: Ed25519 private key bytes (32 bytes)
        encrypt_key: AES-128 key bytes (16 bytes)
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Raises:
        ValueError: If keys are invalid
        Claim169Exception: If encoding fails
    """
    ...

def encode_unsigned(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential without signature (INSECURE - testing only).

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Raises:
        Claim169Exception: If encoding fails
    """
    ...

def encode_with_signer(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    signer: SignerCallback,
    algorithm: SignAlgorithm,
    key_id: Optional[bytes] = None,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with a custom signer callback.

    Use for integration with external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Remote signing services
    - Smart cards and TPMs

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        signer: Callable that signs data
        algorithm: Signing algorithm ("EdDSA" or "ES256")
        key_id: Optional key identifier bytes
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Example:
        def my_sign(algorithm, key_id, data):
            return crypto_provider.sign(key_id, data)

        qr_text = encode_with_signer(claim, meta, my_sign, "EdDSA")
    """
    ...

def encode_with_signer_and_encryptor(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    signer: SignerCallback,
    sign_algorithm: SignAlgorithm,
    encryptor: EncryptorCallback,
    encrypt_algorithm: EncryptAlgorithm,
    key_id: Optional[bytes] = None,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode a Claim 169 credential with custom signer and encryptor callbacks.

    Use for integration with external crypto providers such as:
    - Hardware Security Modules (HSMs)
    - Cloud KMS (AWS KMS, Google Cloud KMS, Azure Key Vault)
    - Remote signing services

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        signer: Callable that signs data
        sign_algorithm: Signing algorithm ("EdDSA" or "ES256")
        encryptor: Callable that encrypts data
        encrypt_algorithm: Encryption algorithm ("A256GCM" or "A128GCM")
        key_id: Optional key identifier bytes
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation

    Example:
        def my_sign(algorithm, key_id, data):
            return crypto_provider.sign(key_id, data)

        def my_encrypt(algorithm, key_id, nonce, aad, plaintext):
            return crypto_provider.encrypt(key_id, nonce, aad, plaintext)

        qr_text = encode_with_signer_and_encryptor(
            claim, meta, my_sign, "EdDSA", my_encrypt, "A256GCM"
        )
    """
    ...

def encode_with_encryptor(
    claim169: Claim169Input,
    cwt_meta: CwtMetaInput,
    sign_key: bytes,
    encryptor: EncryptorCallback,
    encrypt_algorithm: EncryptAlgorithm,
    skip_biometrics: bool = False,
) -> str:
    """
    Encode with software signing and custom encryptor callback.

    Args:
        claim169: Identity data to encode
        cwt_meta: CWT metadata (issuer, expiration, etc.)
        sign_key: Ed25519 private key bytes (32 bytes)
        encryptor: Callable that encrypts data
        encrypt_algorithm: Encryption algorithm ("A256GCM" or "A128GCM")
        skip_biometrics: If True, exclude biometric data to reduce QR size

    Returns:
        Base45-encoded string suitable for QR code generation
    """
    ...

def generate_nonce() -> bytes:
    """
    Generate a cryptographically secure random 12-byte nonce for AES-GCM encryption.

    Returns:
        12-byte nonce suitable for AES-GCM IV
    """
    ...



"""Tests for the claim169 encode functionality."""

import pytest

import claim169


def create_claim(
    id=None,
    full_name=None,
    version=None,
    language=None,
    first_name=None,
    middle_name=None,
    last_name=None,
    date_of_birth=None,
    gender=None,
    address=None,
    email=None,
    phone=None,
    nationality=None,
    marital_status=None,
    guardian=None,
    photo=None,
    photo_format=None,
    secondary_full_name=None,
    secondary_language=None,
    location_code=None,
    legal_status=None,
    country_of_issuance=None,
):
    """Helper to create Claim169Input with all optional fields."""
    claim = claim169.Claim169Input(id=id, full_name=full_name)
    if version is not None:
        claim.version = version
    if language is not None:
        claim.language = language
    if first_name is not None:
        claim.first_name = first_name
    if middle_name is not None:
        claim.middle_name = middle_name
    if last_name is not None:
        claim.last_name = last_name
    if date_of_birth is not None:
        claim.date_of_birth = date_of_birth
    if gender is not None:
        claim.gender = gender
    if address is not None:
        claim.address = address
    if email is not None:
        claim.email = email
    if phone is not None:
        claim.phone = phone
    if nationality is not None:
        claim.nationality = nationality
    if marital_status is not None:
        claim.marital_status = marital_status
    if guardian is not None:
        claim.guardian = guardian
    if photo is not None:
        claim.photo = photo
    if photo_format is not None:
        claim.photo_format = photo_format
    if secondary_full_name is not None:
        claim.secondary_full_name = secondary_full_name
    if secondary_language is not None:
        claim.secondary_language = secondary_language
    if location_code is not None:
        claim.location_code = location_code
    if legal_status is not None:
        claim.legal_status = legal_status
    if country_of_issuance is not None:
        claim.country_of_issuance = country_of_issuance
    return claim


def create_meta(
    issuer=None,
    subject=None,
    expires_at=None,
    not_before=None,
    issued_at=None,
):
    """Helper to create CwtMetaInput with all optional fields."""
    meta = claim169.CwtMetaInput()
    if issuer is not None:
        meta.issuer = issuer
    if subject is not None:
        meta.subject = subject
    if expires_at is not None:
        meta.expires_at = expires_at
    if not_before is not None:
        meta.not_before = not_before
    if issued_at is not None:
        meta.issued_at = issued_at
    return meta


class TestGenerateNonce:
    """Tests for nonce generation."""

    def test_generate_nonce_returns_bytes(self):
        """Test that generate_nonce returns bytes or list."""
        nonce = claim169.generate_nonce()
        # PyO3 returns list, convert if needed
        if isinstance(nonce, list):
            nonce = bytes(nonce)
        assert isinstance(nonce, bytes)

    def test_generate_nonce_length(self):
        """Test that generate_nonce returns 12 bytes (AES-GCM standard)."""
        nonce = claim169.generate_nonce()
        if isinstance(nonce, list):
            nonce = bytes(nonce)
        assert len(nonce) == 12

    def test_generate_nonce_unique(self):
        """Test that consecutive nonces are unique."""
        nonces = []
        for _ in range(100):
            n = claim169.generate_nonce()
            if isinstance(n, list):
                n = bytes(n)
            nonces.append(n)
        # All nonces should be unique
        assert len(set(nonces)) == 100


class TestClaim169Input:
    """Tests for Claim169Input construction."""

    def test_create_minimal(self):
        """Test creating a minimal Claim169Input."""
        claim = claim169.Claim169Input(
            id="TEST-001",
            full_name="Test Person"
        )
        assert claim.id == "TEST-001"
        assert claim.full_name == "Test Person"

    def test_create_with_all_demographics(self):
        """Test creating Claim169Input with all demographic fields."""
        claim = create_claim(
            id="TEST-002",
            version="1.0.0",
            language="en",
            full_name="Test Person",
            first_name="Test",
            middle_name="Middle",
            last_name="Person",
            date_of_birth="1990-01-15",
            gender=1,  # Male
            address="123 Test Street",
            email="test@example.com",
            phone="+1234567890",
            nationality="US",
            marital_status=1,  # Single
            guardian="Guardian Name",
            secondary_full_name="Alternative Name",
            secondary_language="es",
            location_code="US-CA",
            legal_status="citizen",
            country_of_issuance="US"
        )
        assert claim.id == "TEST-002"
        assert claim.full_name == "Test Person"
        assert claim.gender == 1


class TestCwtMetaInput:
    """Tests for CwtMetaInput construction."""

    def test_create_minimal(self):
        """Test creating a minimal CwtMetaInput."""
        meta = claim169.CwtMetaInput()
        # Should not raise

    def test_create_with_all_fields(self):
        """Test creating CwtMetaInput with all fields."""
        meta = create_meta(
            issuer="https://example.org",
            subject="user123",
            expires_at=1800000000,
            not_before=1700000000,
            issued_at=1700000000
        )
        assert meta.issuer == "https://example.org"
        assert meta.expires_at == 1800000000


class TestEncodeWithEd25519:
    """Tests for Ed25519 encoding."""

    def test_encode_minimal(self, ed25519_signed_vector):
        """Test encoding a minimal claim with Ed25519."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )

        claim = create_claim(
            id="ENCODE-TEST-001",
            full_name="Encode Test Person"
        )
        meta = create_meta(
            issuer="https://test.example.org",
            issued_at=1700000000,
            expires_at=1900000000  # Far future to avoid expiration
        )

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)

        assert isinstance(qr_data, str)
        assert len(qr_data) > 0

    def test_encode_decode_roundtrip(self, ed25519_signed_vector):
        """Test that encode->decode roundtrip preserves data."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        original_id = "ROUNDTRIP-001"
        original_name = "Roundtrip Test Person"
        original_email = "roundtrip@test.org"

        claim = create_claim(
            id=original_id,
            full_name=original_name,
            email=original_email
        )
        meta = create_meta(
            issuer="https://roundtrip.example.org",
            issued_at=1700000000,
            expires_at=1900000000
        )

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        assert result.claim169.id == original_id
        assert result.claim169.full_name == original_name
        assert result.claim169.email == original_email
        assert result.cwt_meta.issuer == "https://roundtrip.example.org"
        assert result.verification_status == "verified"

    def test_invalid_key_length(self):
        """Test that invalid key length raises error."""
        claim = claim169.Claim169Input(id="TEST", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises((ValueError, claim169.Claim169Exception)):
            claim169.encode_with_ed25519(claim, meta, bytes(16))  # Too short

    def test_encode_with_all_demographics(self, ed25519_signed_vector):
        """Test encoding with all demographic fields."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        claim = create_claim(
            id="FULL-DEMO-001",
            version="1.0.0",
            language="en",
            full_name="Full Demographics Person",
            first_name="Full",
            middle_name="Demo",
            last_name="Person",
            date_of_birth="1985-06-15",
            gender=2,  # Female
            address="456 Demo Avenue, Test City",
            email="full@demo.org",
            phone="+1987654321",
            nationality="CA",
            marital_status=2,  # Married
            secondary_full_name="Nom Complet",
            secondary_language="fr",
            location_code="CA-QC",
            legal_status="permanent_resident",
            country_of_issuance="CA"
        )
        meta = create_meta(
            issuer="https://demographics.example.org",
            subject="demo-subject",
            issued_at=1700000000,
            expires_at=1900000000,
            not_before=1700000000
        )

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        assert result.claim169.id == "FULL-DEMO-001"
        assert result.claim169.version == "1.0.0"
        assert result.claim169.language == "en"
        assert result.claim169.full_name == "Full Demographics Person"
        assert result.claim169.first_name == "Full"
        assert result.claim169.middle_name == "Demo"
        assert result.claim169.last_name == "Person"
        assert result.claim169.date_of_birth == "1985-06-15"
        assert result.claim169.gender == 2
        assert result.claim169.address == "456 Demo Avenue, Test City"
        assert result.claim169.email == "full@demo.org"
        assert result.claim169.phone == "+1987654321"
        assert result.claim169.nationality == "CA"
        assert result.claim169.marital_status == 2
        assert result.claim169.secondary_full_name == "Nom Complet"
        assert result.claim169.secondary_language == "fr"
        assert result.claim169.location_code == "CA-QC"
        assert result.claim169.legal_status == "permanent_resident"
        assert result.claim169.country_of_issuance == "CA"
        assert result.verification_status == "verified"


class TestEncodeWithEcdsaP256:
    """Tests for ECDSA P-256 encoding."""

    def test_encode_decode_roundtrip(self):
        """Test ECDSA P-256 encode->decode roundtrip."""
        # Generate a test key pair using cryptography library
        from cryptography.hazmat.primitives.asymmetric import ec
        from cryptography.hazmat.backends import default_backend

        private_key_obj = ec.generate_private_key(ec.SECP256R1(), default_backend())
        private_key_bytes = private_key_obj.private_numbers().private_value.to_bytes(32, 'big')
        public_key_obj = private_key_obj.public_key()
        public_numbers = public_key_obj.public_numbers()
        # Uncompressed SEC1 format: 0x04 || x || y
        public_key_bytes = b'\x04' + public_numbers.x.to_bytes(32, 'big') + public_numbers.y.to_bytes(32, 'big')

        original_id = "ECDSA-ROUNDTRIP-001"
        original_name = "ECDSA Roundtrip Person"

        claim = create_claim(
            id=original_id,
            full_name=original_name
        )
        meta = create_meta(
            issuer="https://ecdsa.example.org",
            issued_at=1700000000,
            expires_at=1900000000
        )

        qr_data = claim169.encode_with_ecdsa_p256(claim, meta, private_key_bytes)
        result = claim169.decode_with_ecdsa_p256(qr_data, public_key_bytes)

        assert result.claim169.id == original_id
        assert result.claim169.full_name == original_name
        assert result.verification_status == "verified"

    def test_invalid_key_length(self):
        """Test that invalid key length raises error."""
        claim = claim169.Claim169Input(id="TEST", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises((ValueError, claim169.Claim169Exception)):
            claim169.encode_with_ecdsa_p256(claim, meta, bytes(16))  # Too short


class TestEncodeSignedEncrypted:
    """Tests for signed and encrypted encoding."""

    def test_encode_decode_roundtrip(self, encrypted_signed_vector):
        """Test signed+encrypted encode->decode roundtrip."""
        sign_private_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["private_key_hex"]
        )
        sign_public_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["public_key_hex"]
        )
        encrypt_key = bytes.fromhex(
            encrypted_signed_vector["encryption_key"]["symmetric_key_hex"]
        )

        original_id = "ENC-SIGN-ROUNDTRIP-001"
        original_name = "Encrypted Signed Roundtrip"

        claim = create_claim(
            id=original_id,
            full_name=original_name
        )
        meta = create_meta(
            issuer="https://encrypted.example.org",
            issued_at=1700000000,
            expires_at=1900000000
        )

        qr_data = claim169.encode_signed_encrypted(
            claim, meta, sign_private_key, encrypt_key
        )

        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

        public_key = Ed25519PublicKey.from_public_bytes(sign_public_key)

        def verify_callback(algorithm, key_id, data, signature):
            public_key.verify(bytes(signature), bytes(data))

        result = claim169.decode_encrypted_aes(
            qr_data,
            encrypt_key,
            verifier=verify_callback
        )

        assert result.claim169.id == original_id
        assert result.claim169.full_name == original_name
        assert result.verification_status == "verified"

    def test_decryption_with_wrong_key_fails(self, encrypted_signed_vector):
        """Test that decryption with wrong key fails."""
        sign_private_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["private_key_hex"]
        )
        encrypt_key = bytes.fromhex(
            encrypted_signed_vector["encryption_key"]["symmetric_key_hex"]
        )
        wrong_decrypt_key = bytes(32)  # All zeros

        claim = claim169.Claim169Input(id="TEST", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_data = claim169.encode_signed_encrypted(
            claim, meta, sign_private_key, encrypt_key
        )

        with pytest.raises(claim169.Claim169Exception):
            claim169.decode_encrypted_aes(qr_data, wrong_decrypt_key, allow_unverified=True)

    def test_invalid_sign_key_length(self, encrypted_signed_vector):
        """Test that invalid signing key length raises error."""
        encrypt_key = bytes.fromhex(
            encrypted_signed_vector["encryption_key"]["symmetric_key_hex"]
        )
        claim = claim169.Claim169Input(id="TEST", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises((ValueError, claim169.Claim169Exception)):
            claim169.encode_signed_encrypted(claim, meta, bytes(16), encrypt_key)

    def test_invalid_encrypt_key_length(self, encrypted_signed_vector):
        """Test that invalid encryption key length raises error."""
        sign_private_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["private_key_hex"]
        )
        claim = claim169.Claim169Input(id="TEST", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises((ValueError, claim169.Claim169Exception)):
            claim169.encode_signed_encrypted(claim, meta, sign_private_key, bytes(16))


class TestEncodeUnsigned:
    """Tests for unsigned encoding (testing only)."""

    def test_encode_decode_roundtrip(self):
        """Test unsigned encode->decode roundtrip."""
        original_id = "UNSIGNED-001"
        original_name = "Unsigned Test Person"

        claim = create_claim(
            id=original_id,
            full_name=original_name
        )
        meta = create_meta(
            issuer="https://unsigned.example.org",
            issued_at=1700000000,
            expires_at=1900000000
        )

        qr_data = claim169.encode_unsigned(claim, meta)
        result = claim169.decode_unverified(qr_data)

        assert result.claim169.id == original_id
        assert result.claim169.full_name == original_name
        assert result.verification_status == "skipped"

    def test_produces_smaller_output(self, ed25519_signed_vector):
        """Test that unsigned encoding produces smaller output than signed."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )

        claim = create_claim(
            id="SIZE-TEST",
            full_name="Size Test Person"
        )
        meta = create_meta(
            issuer="https://size.example.org",
            expires_at=1900000000
        )

        signed_qr = claim169.encode_with_ed25519(claim, meta, private_key)
        unsigned_qr = claim169.encode_unsigned(claim, meta)

        # Unsigned should be smaller (no signature overhead)
        assert len(unsigned_qr) < len(signed_qr)


class TestEncoderEdgeCases:
    """Edge case tests for encoder."""

    def test_empty_optional_fields(self, ed25519_signed_vector):
        """Test encoding with None for all optional fields."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        # Only required fields
        claim = claim169.Claim169Input()
        meta = create_meta(expires_at=1900000000)

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        # All optional fields should be None
        assert result.claim169.id is None
        assert result.claim169.full_name is None
        assert result.claim169.email is None

    def test_unicode_in_fields(self, ed25519_signed_vector):
        """Test encoding with Unicode characters."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        unicode_name = "日本語テスト"
        unicode_address = "东京都渋谷区"

        claim = create_claim(
            id="UNICODE-001",
            full_name=unicode_name,
            address=unicode_address
        )
        meta = create_meta(expires_at=1900000000)

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        assert result.claim169.full_name == unicode_name
        assert result.claim169.address == unicode_address

    def test_large_field_values(self, ed25519_signed_vector):
        """Test encoding with large field values."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        # Long but reasonable values
        long_name = "A" * 200
        long_address = "B" * 500

        claim = create_claim(
            id="LARGE-001",
            full_name=long_name,
            address=long_address
        )
        meta = create_meta(expires_at=1900000000)

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        assert result.claim169.full_name == long_name
        assert result.claim169.address == long_address

    def test_special_characters_in_fields(self, ed25519_signed_vector):
        """Test encoding with special characters."""
        private_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["private_key_hex"]
        )
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )

        special_email = "test+special@example.org"
        special_phone = "+1 (234) 567-8900"

        claim = create_claim(
            id="SPECIAL-001",
            full_name="O'Brien-Smith",
            email=special_email,
            phone=special_phone
        )
        meta = create_meta(expires_at=1900000000)

        qr_data = claim169.encode_with_ed25519(claim, meta, private_key)
        result = claim169.decode_with_ed25519(qr_data, public_key)

        assert result.claim169.full_name == "O'Brien-Smith"
        assert result.claim169.email == special_email
        assert result.claim169.phone == special_phone



"""Tests for custom crypto provider functions (signer, encryptor callbacks)."""

import secrets

import pytest

import claim169


class TestEncodeWithSigner:
    """Tests for encode_with_signer function."""

    def test_encode_with_custom_signer_eddsa(self):
        """Test encoding with a custom EdDSA signer callback."""
        # Generate a test Ed25519 keypair using software crypto
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

        private_key = Ed25519PrivateKey.generate()
        public_key_bytes = private_key.public_key().public_bytes_raw()

        def my_signer(algorithm, key_id, data):
            assert algorithm == "EdDSA"
            return private_key.sign(data)

        claim = claim169.Claim169Input(id="custom-signer-test", full_name="Test User")
        meta = claim169.CwtMetaInput(issuer="test")

        # Encode with custom signer
        qr_text = claim169.encode_with_signer(claim, meta, my_signer, "EdDSA")
        assert qr_text

        # Verify we can decode it
        result = claim169.decode_with_ed25519(
            qr_text, public_key_bytes, validate_timestamps=False
        )
        assert result.claim169.id == "custom-signer-test"
        assert result.verification_status == "verified"

    def test_encode_with_custom_signer_es256(self):
        """Test encoding with a custom ES256 signer callback."""
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import ec
        from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

        private_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        public_key_bytes = private_key.public_key().public_bytes(
            Encoding.X962, PublicFormat.CompressedPoint
        )

        def my_signer(algorithm, key_id, data):
            from cryptography.hazmat.primitives import hashes

            assert algorithm == "ES256"
            sig = private_key.sign(data, ec.ECDSA(hashes.SHA256()))
            # Convert DER to raw r||s format (64 bytes)
            from cryptography.hazmat.primitives.asymmetric.utils import (
                decode_dss_signature,
            )

            r, s = decode_dss_signature(sig)
            return r.to_bytes(32, "big") + s.to_bytes(32, "big")

        claim = claim169.Claim169Input(id="es256-signer", full_name="ES256 User")
        meta = claim169.CwtMetaInput(issuer="test")

        qr_text = claim169.encode_with_signer(claim, meta, my_signer, "ES256")
        assert qr_text

        result = claim169.decode_with_ecdsa_p256(
            qr_text, public_key_bytes, validate_timestamps=False
        )
        assert result.claim169.id == "es256-signer"
        assert result.verification_status == "verified"

    def test_encode_with_signer_with_key_id(self):
        """Test that key_id is passed to the signer callback."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

        private_key = Ed25519PrivateKey.generate()
        public_key_bytes = private_key.public_key().public_bytes_raw()
        expected_key_id = b"my-key-123"
        received_key_id = None

        def my_signer(algorithm, key_id, data):
            nonlocal received_key_id
            received_key_id = key_id
            return private_key.sign(data)

        claim = claim169.Claim169Input(id="key-id-test", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_text = claim169.encode_with_signer(
            claim, meta, my_signer, "EdDSA", key_id=expected_key_id
        )
        assert qr_text
        assert received_key_id == expected_key_id

        result = claim169.decode_with_ed25519(
            qr_text, public_key_bytes, validate_timestamps=False
        )
        assert result.claim169.id == "key-id-test"

    def test_encode_with_signer_skip_biometrics(self):
        """Test skip_biometrics parameter with custom signer."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

        private_key = Ed25519PrivateKey.generate()

        def my_signer(algorithm, key_id, data):
            return private_key.sign(data)

        claim = claim169.Claim169Input(id="skip-bio", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_with_bio = claim169.encode_with_signer(
            claim, meta, my_signer, "EdDSA", skip_biometrics=False
        )
        qr_without_bio = claim169.encode_with_signer(
            claim, meta, my_signer, "EdDSA", skip_biometrics=True
        )

        # Both should produce valid output
        assert qr_with_bio
        assert qr_without_bio

    def test_encode_with_signer_invalid_algorithm(self):
        """Test that invalid algorithm is rejected."""

        def my_signer(algorithm, key_id, data):
            return b"fake"

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises(ValueError, match="Unsupported algorithm"):
            claim169.encode_with_signer(claim, meta, my_signer, "RSA256")

    def test_encode_with_signer_callback_error(self):
        """Test that callback errors are propagated."""

        def failing_signer(algorithm, key_id, data):
            raise RuntimeError("Crypto provider unavailable")

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises(claim169.Claim169Exception):
            claim169.encode_with_signer(claim, meta, failing_signer, "EdDSA")


class TestEncodeWithSignerAndEncryptor:
    """Tests for encode_with_signer_and_encryptor function."""

    def test_encode_with_custom_signer_and_encryptor(self):
        """Test encoding with both custom signer and encryptor."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
        from cryptography.hazmat.primitives.ciphers.aead import AESGCM

        # Generate keys
        sign_private = Ed25519PrivateKey.generate()
        sign_public_bytes = sign_private.public_key().public_bytes_raw()
        aes_key = secrets.token_bytes(32)

        def my_signer(algorithm, key_id, data):
            return sign_private.sign(data)

        def my_encryptor(algorithm, key_id, nonce, aad, plaintext):
            assert algorithm == "A256GCM"
            aesgcm = AESGCM(aes_key)
            return aesgcm.encrypt(nonce, plaintext, aad)

        claim = claim169.Claim169Input(id="custom-both", full_name="Custom Crypto")
        meta = claim169.CwtMetaInput(issuer="test")

        qr_text = claim169.encode_with_signer_and_encryptor(
            claim, meta, my_signer, "EdDSA", my_encryptor, "A256GCM"
        )
        assert qr_text

        # Decode with the keys
        def my_verifier(algorithm, key_id, data, signature):
            sign_private.public_key().verify(signature, data)

        def my_decryptor(algorithm, key_id, nonce, aad, ciphertext):
            aesgcm = AESGCM(aes_key)
            return aesgcm.decrypt(nonce, ciphertext, aad)

        result = claim169.decode_with_decryptor(
            qr_text, my_decryptor, verifier=my_verifier
        )
        assert result.claim169.id == "custom-both"
        assert result.verification_status == "verified"

    def test_encode_with_aes128_encryptor(self):
        """Test encoding with AES-128 custom encryptor."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
        from cryptography.hazmat.primitives.ciphers.aead import AESGCM

        sign_private = Ed25519PrivateKey.generate()
        aes_key = secrets.token_bytes(16)  # AES-128

        def my_signer(algorithm, key_id, data):
            return sign_private.sign(data)

        def my_encryptor(algorithm, key_id, nonce, aad, plaintext):
            assert algorithm == "A128GCM"
            aesgcm = AESGCM(aes_key)
            return aesgcm.encrypt(nonce, plaintext, aad)

        claim = claim169.Claim169Input(id="aes128-test", full_name="AES128 User")
        meta = claim169.CwtMetaInput()

        qr_text = claim169.encode_with_signer_and_encryptor(
            claim, meta, my_signer, "EdDSA", my_encryptor, "A128GCM"
        )
        assert qr_text

        # Decode
        result = claim169.decode_encrypted_aes128(qr_text, aes_key, allow_unverified=True)
        assert result.claim169.id == "aes128-test"


class TestEncodeWithEncryptor:
    """Tests for encode_with_encryptor function (software signing + custom encryption)."""

    def test_encode_with_custom_encryptor(self):
        """Test encoding with software signing and custom encryptor."""
        from cryptography.hazmat.primitives.ciphers.aead import AESGCM

        sign_key = secrets.token_bytes(32)
        aes_key = secrets.token_bytes(32)

        def my_encryptor(algorithm, key_id, nonce, aad, plaintext):
            aesgcm = AESGCM(aes_key)
            return aesgcm.encrypt(nonce, plaintext, aad)

        claim = claim169.Claim169Input(id="custom-enc", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_text = claim169.encode_with_encryptor(
            claim, meta, sign_key, my_encryptor, "A256GCM"
        )
        assert qr_text

        # Decode with software decryption
        result = claim169.decode_encrypted_aes(qr_text, aes_key, allow_unverified=True)
        assert result.claim169.id == "custom-enc"


class TestAes128Functions:
    """Tests for AES-128 specific functions."""

    def test_encode_signed_encrypted_aes128(self):
        """Test encode_signed_encrypted_aes128 function."""
        sign_key = secrets.token_bytes(32)
        encrypt_key = secrets.token_bytes(16)  # AES-128

        claim = claim169.Claim169Input(id="aes128-enc", full_name="AES128")
        meta = claim169.CwtMetaInput(issuer="test")

        qr_text = claim169.encode_signed_encrypted_aes128(
            claim, meta, sign_key, encrypt_key
        )
        assert qr_text

        result = claim169.decode_encrypted_aes128(
            qr_text, encrypt_key, allow_unverified=True
        )
        assert result.claim169.id == "aes128-enc"

    def test_decode_encrypted_aes128_key_validation(self):
        """Test that AES-128 decryption validates key length."""
        # Create an encrypted credential with AES-256 (for testing)
        sign_key = secrets.token_bytes(32)
        encrypt_key = secrets.token_bytes(32)  # AES-256

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_text = claim169.encode_signed_encrypted(claim, meta, sign_key, encrypt_key)

        # Try to decrypt with wrong key size
        wrong_key = secrets.token_bytes(16)  # AES-128 key
        with pytest.raises(claim169.DecryptionError, match="does not match key type"):
            # This should fail because AES-128 key can't decrypt AES-256 credential
            claim169.decode_encrypted_aes128(qr_text, wrong_key, allow_unverified=True)

    def test_decode_encrypted_aes256_key_validation(self):
        """Test that AES-256 decryption validates key length."""
        sign_key = secrets.token_bytes(32)
        encrypt_key = secrets.token_bytes(16)  # AES-128

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        qr_text = claim169.encode_signed_encrypted_aes128(
            claim, meta, sign_key, encrypt_key
        )

        # Try to decrypt with wrong key size
        wrong_key = secrets.token_bytes(32)  # AES-256 key
        with pytest.raises(claim169.DecryptionError, match="does not match key type"):
            # This should fail because AES-256 key can't decrypt AES-128 credential
            claim169.decode_encrypted_aes256(qr_text, wrong_key, allow_unverified=True)


class TestRoundtripCustomCrypto:
    """Roundtrip tests with custom crypto providers."""

    def test_full_roundtrip_custom_signer_and_verifier(self):
        """Test encode with custom signer and decode with custom verifier."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

        private_key = Ed25519PrivateKey.generate()

        def my_signer(algorithm, key_id, data):
            return private_key.sign(data)

        def my_verifier(algorithm, key_id, data, signature):
            private_key.public_key().verify(signature, data)

        claim = claim169.Claim169Input(id="roundtrip-test", full_name="Roundtrip User")
        claim.email = "roundtrip@test.com"
        claim.gender = 1
        meta = claim169.CwtMetaInput(issuer="https://issuer.example")

        # Encode
        qr_text = claim169.encode_with_signer(claim, meta, my_signer, "EdDSA")

        # Decode
        result = claim169.decode_with_verifier(qr_text, my_verifier)

        assert result.claim169.id == "roundtrip-test"
        assert result.claim169.full_name == "Roundtrip User"
        assert result.claim169.email == "roundtrip@test.com"
        assert result.claim169.gender == 1
        assert result.cwt_meta.issuer == "https://issuer.example"
        assert result.verification_status == "verified"

    def test_full_roundtrip_custom_crypto_all(self):
        """Test full roundtrip with custom signer, encryptor, verifier, decryptor."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
        from cryptography.hazmat.primitives.ciphers.aead import AESGCM

        private_key = Ed25519PrivateKey.generate()
        aes_key = secrets.token_bytes(32)

        def my_signer(algorithm, key_id, data):
            return private_key.sign(data)

        def my_encryptor(algorithm, key_id, nonce, aad, plaintext):
            return AESGCM(aes_key).encrypt(nonce, plaintext, aad)

        def my_verifier(algorithm, key_id, data, signature):
            private_key.public_key().verify(signature, data)

        def my_decryptor(algorithm, key_id, nonce, aad, ciphertext):
            return AESGCM(aes_key).decrypt(nonce, ciphertext, aad)

        claim = claim169.Claim169Input(id="full-custom", full_name="Full Custom Test")
        meta = claim169.CwtMetaInput(issuer="https://full-custom.example")

        # Encode with custom signer and encryptor
        qr_text = claim169.encode_with_signer_and_encryptor(
            claim, meta, my_signer, "EdDSA", my_encryptor, "A256GCM"
        )

        # Decode with custom decryptor and verifier
        result = claim169.decode_with_decryptor(
            qr_text, my_decryptor, verifier=my_verifier
        )

        assert result.claim169.id == "full-custom"
        assert result.claim169.full_name == "Full Custom Test"
        assert result.cwt_meta.issuer == "https://full-custom.example"
        assert result.verification_status == "verified"


class TestErrorHandling:
    """Test error handling in custom crypto providers."""

    def test_signer_returns_wrong_type(self):
        """Test that returning wrong type from signer raises error."""

        def bad_signer(algorithm, key_id, data):
            return "not bytes"  # Should return bytes

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises(claim169.Claim169Exception):
            claim169.encode_with_signer(claim, meta, bad_signer, "EdDSA")

    def test_encryptor_returns_wrong_type(self):
        """Test that returning wrong type from encryptor raises error."""
        sign_key = secrets.token_bytes(32)

        def bad_encryptor(algorithm, key_id, nonce, aad, plaintext):
            return 12345  # Should return bytes

        claim = claim169.Claim169Input(id="test", full_name="Test")
        meta = claim169.CwtMetaInput()

        with pytest.raises(claim169.Claim169Exception):
            claim169.encode_with_encryptor(
                claim, meta, sign_key, bad_encryptor, "A256GCM"
            )



"""Pytest configuration and fixtures for claim169 tests."""

import json
from pathlib import Path

import pytest


TEST_VECTORS_PATH = Path(__file__).parent.parent.parent.parent / "test-vectors"


@pytest.fixture
def test_vectors_path() -> Path:
    """Return the path to the test vectors directory."""
    return TEST_VECTORS_PATH


def load_test_vector(category: str, name: str) -> dict:
    """Load a test vector JSON file."""
    file_path = TEST_VECTORS_PATH / category / f"{name}.json"
    with open(file_path) as f:
        return json.load(f)


@pytest.fixture
def minimal_vector() -> dict:
    """Load the minimal test vector."""
    return load_test_vector("valid", "minimal")


@pytest.fixture
def demographics_full_vector() -> dict:
    """Load the demographics-full test vector."""
    return load_test_vector("valid", "demographics-full")


@pytest.fixture
def with_face_vector() -> dict:
    """Load the with-face test vector."""
    return load_test_vector("valid", "with-face")


@pytest.fixture
def with_fingerprints_vector() -> dict:
    """Load the with-fingerprints test vector."""
    return load_test_vector("valid", "with-fingerprints")


@pytest.fixture
def ed25519_signed_vector() -> dict:
    """Load the ed25519-signed test vector."""
    return load_test_vector("valid", "ed25519-signed")


@pytest.fixture
def encrypted_aes256_vector() -> dict:
    """Load the encrypted-aes256 test vector."""
    return load_test_vector("valid", "encrypted-aes256")


@pytest.fixture
def encrypted_signed_vector() -> dict:
    """Load the encrypted-signed test vector."""
    return load_test_vector("valid", "encrypted-signed")


@pytest.fixture
def ecdsa_p256_signed_vector() -> dict:
    """Load the ecdsa-p256-signed test vector."""
    return load_test_vector("valid", "ecdsa-p256-signed")


@pytest.fixture
def with_all_biometrics_vector() -> dict:
    """Load the with-all-biometrics test vector."""
    return load_test_vector("valid", "with-all-biometrics")


@pytest.fixture
def bad_base45_vector() -> dict:
    """Load the bad-base45 test vector."""
    return load_test_vector("invalid", "bad-base45")


@pytest.fixture
def bad_zlib_vector() -> dict:
    """Load the bad-zlib test vector."""
    return load_test_vector("invalid", "bad-zlib")


@pytest.fixture
def not_cose_vector() -> dict:
    """Load the not-cose test vector."""
    return load_test_vector("invalid", "not-cose")


@pytest.fixture
def missing_169_vector() -> dict:
    """Load the missing-169 test vector."""
    return load_test_vector("invalid", "missing-169")


@pytest.fixture
def expired_vector() -> dict:
    """Load the expired test vector."""
    return load_test_vector("edge", "expired")


@pytest.fixture
def unknown_fields_vector() -> dict:
    """Load the unknown-fields test vector."""
    return load_test_vector("edge", "unknown-fields")


@pytest.fixture
def not_yet_valid_vector() -> dict:
    """Load the not-yet-valid test vector."""
    return load_test_vector("edge", "not-yet-valid")


@pytest.fixture
def claim169_example_vector() -> dict:
    """Load the claim169-example test vector (from claim_169.md)."""
    return load_test_vector("valid", "claim169-example")



"""Tests for the claim169 decode functionality."""

import pytest

import claim169


class TestModule:
    """Tests for the module itself."""

    def test_version_returns_string(self):
        """Test that version() returns a version string."""
        version = claim169.version()
        assert isinstance(version, str)
        assert version.count(".") >= 2  # Should be semver format

    def test_dunder_version_matches_version(self):
        """Test that __version__ matches version()."""
        assert claim169.__version__ == claim169.version()

    def test_decode_requires_verification_by_default(self, minimal_vector):
        """Test that decode() requires a key unless allow_unverified=True."""
        with pytest.raises(ValueError):
            claim169.decode(minimal_vector["qr_data"])

        result = claim169.decode(minimal_vector["qr_data"], allow_unverified=True)
        assert result.verification_status == "skipped"


class TestDecodeValidVectors:
    """Tests for decoding valid test vectors."""

    def test_decode_minimal(self, minimal_vector):
        """Test decoding the minimal vector."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])

        expected = minimal_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")
        assert result.verification_status == "skipped"

    def test_decode_demographics_full(self, demographics_full_vector):
        """Test decoding the demographics-full vector."""
        result = claim169.decode_unverified(demographics_full_vector["qr_data"])

        expected = demographics_full_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")
        if "firstName" in expected:
            assert result.claim169.first_name == expected["firstName"]

        expected_meta = demographics_full_vector.get("expected_cwt_meta", {})
        if "issuer" in expected_meta:
            assert result.cwt_meta.issuer == expected_meta["issuer"]

    def test_decode_with_face(self, with_face_vector):
        """Test decoding the with-face vector."""
        result = claim169.decode_unverified(with_face_vector["qr_data"])

        expected = with_face_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.face is not None
        assert len(result.claim169.face) > 0

    def test_decode_with_fingerprints(self, with_fingerprints_vector):
        """Test decoding the with-fingerprints vector."""
        result = claim169.decode_unverified(with_fingerprints_vector["qr_data"])

        expected = with_fingerprints_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")

        # Check at least one fingerprint field is present
        has_fingerprint = any([
            result.claim169.right_thumb is not None,
            result.claim169.right_pointer_finger is not None,
            result.claim169.left_thumb is not None,
        ])
        assert has_fingerprint

    def test_decode_claim169_example(self, claim169_example_vector):
        """Test decoding the example from claim_169.md specification."""
        result = claim169.decode_unverified(claim169_example_vector["qr_data"])

        expected = claim169_example_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")
        assert result.claim169.version == expected.get("version")
        assert result.claim169.language == expected.get("language")
        assert result.claim169.date_of_birth == expected.get("dateOfBirth")
        assert result.claim169.gender == expected.get("gender")
        assert result.claim169.address == expected.get("address")
        assert result.claim169.email == expected.get("email")
        assert result.claim169.phone == expected.get("phone")
        assert result.claim169.nationality == expected.get("nationality")
        assert result.claim169.marital_status == expected.get("maritalStatus")
        assert result.claim169.secondary_full_name == expected.get("secondaryFullName")
        assert result.claim169.secondary_language == expected.get("secondaryLanguage")
        assert result.claim169.location_code == expected.get("locationCode")
        assert result.claim169.legal_status == expected.get("legalStatus")
        assert result.claim169.country_of_issuance == expected.get("countryOfIssuance")

        # Check face biometric is present
        assert result.claim169.face is not None
        assert len(result.claim169.face) > 0
        assert result.claim169.face[0].format == expected["face"][0]["format"]
        assert result.claim169.face[0].sub_format == expected["face"][0]["subFormat"]

        # Check CWT metadata matches example
        expected_meta = claim169_example_vector.get("expected_cwt_meta", {})
        assert result.cwt_meta.issuer == expected_meta.get("issuer")
        assert result.cwt_meta.issued_at == expected_meta.get("issuedAt")
        assert result.cwt_meta.expires_at == expected_meta.get("expiresAt")
        assert result.cwt_meta.not_before == expected_meta.get("notBefore")


class TestDecodeInvalidVectors:
    """Tests for rejecting invalid test vectors."""

    def test_reject_bad_base45(self, bad_base45_vector):
        """Test that bad-base45 vector is rejected."""
        with pytest.raises(claim169.Base45DecodeError):
            claim169.decode_unverified(bad_base45_vector["qr_data"])

    def test_reject_bad_zlib(self, bad_zlib_vector):
        """Test that bad-zlib vector is rejected."""
        with pytest.raises(claim169.DecompressError):
            claim169.decode_unverified(bad_zlib_vector["qr_data"])

    def test_reject_not_cose(self, not_cose_vector):
        """Test that not-cose vector is rejected."""
        with pytest.raises(claim169.CoseParseError):
            claim169.decode_unverified(not_cose_vector["qr_data"])

    def test_reject_missing_169(self, missing_169_vector):
        """Test that missing-169 vector is rejected."""
        with pytest.raises(claim169.Claim169NotFoundError):
            claim169.decode_unverified(missing_169_vector["qr_data"])


class TestDecodeEdgeCases:
    """Tests for edge case vectors."""

    def test_handle_unknown_fields(self, unknown_fields_vector):
        """Test that unknown fields are handled gracefully."""
        result = claim169.decode_unverified(unknown_fields_vector["qr_data"])

        expected = unknown_fields_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")

    def test_expired_token_rejected(self, expired_vector):
        """Test that expired tokens are rejected when timestamp validation is enabled."""
        # The library enforces timestamp validation and rejects expired tokens
        with pytest.raises(claim169.Claim169Exception) as exc_info:
            claim169.decode_unverified(expired_vector["qr_data"])

        # Verify the error message indicates expiration
        expected_meta = expired_vector["expected_cwt_meta"]
        assert "expired" in str(exc_info.value).lower()
        assert str(expected_meta["expiresAt"]) in str(exc_info.value)

    def test_not_yet_valid_token_rejected(self, not_yet_valid_vector):
        """Test that not-yet-valid tokens are rejected when timestamp validation is enabled."""
        # The library enforces timestamp validation and rejects tokens with nbf in the future
        with pytest.raises(claim169.Claim169Exception) as exc_info:
            claim169.decode_unverified(not_yet_valid_vector["qr_data"])

        # Verify the error message indicates the token is not yet valid
        expected_meta = not_yet_valid_vector["expected_cwt_meta"]
        assert "not valid" in str(exc_info.value).lower()
        assert str(expected_meta["notBefore"]) in str(exc_info.value)


class TestDecodeOptions:
    """Tests for decode options."""

    def test_skip_biometrics(self, with_face_vector):
        """Test that skip_biometrics option works."""
        result = claim169.decode_unverified(with_face_vector["qr_data"], skip_biometrics=True)

        expected = with_face_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.face is None

    def test_max_decompressed_bytes_too_small(self, minimal_vector):
        """Test that max_decompressed_bytes limit is enforced."""
        with pytest.raises(claim169.DecompressError):
            claim169.decode_unverified(minimal_vector["qr_data"], max_decompressed_bytes=10)


class TestDecodeWithEd25519:
    """Tests for Ed25519 signature verification."""

    def test_verify_valid_signature(self, ed25519_signed_vector):
        """Test verifying a valid Ed25519 signature."""
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )
        result = claim169.decode_with_ed25519(
            ed25519_signed_vector["qr_data"],
            public_key
        )

        expected = ed25519_signed_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")
        assert result.verification_status == "verified"

    def test_reject_wrong_key(self, ed25519_signed_vector):
        """Test that wrong key fails verification."""
        # Use a non-zero key to avoid "weak key" rejection
        wrong_key = bytes([0x01] + [0x00] * 31)
        with pytest.raises((claim169.Claim169Exception, ValueError)):
            claim169.decode_with_ed25519(
                ed25519_signed_vector["qr_data"],
                wrong_key
            )


class TestDecodeEncrypted:
    """Tests for encrypted payload decoding."""

    def test_decrypt_requires_verifier_by_default(self, encrypted_aes256_vector):
        """Test that encrypted decode requires a verifier unless allow_unverified=True."""
        enc_key = bytes(32)  # length-valid key
        with pytest.raises(ValueError):
            claim169.decode_encrypted_aes(encrypted_aes256_vector["qr_data"], enc_key)

    def test_decrypt_signed_aes256(self, encrypted_signed_vector):
        """Test decrypting an AES-256-GCM encrypted payload containing a signed CWT."""
        enc_key = bytes.fromhex(
            encrypted_signed_vector["encryption_key"]["symmetric_key_hex"]
        )
        sign_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["public_key_hex"]
        )

        # Provide a verifier using the signing key from the test vector
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

        public_key = Ed25519PublicKey.from_public_bytes(sign_key)

        def verify_callback(algorithm, key_id, data, signature):
            public_key.verify(bytes(signature), bytes(data))

        result = claim169.decode_encrypted_aes(
            encrypted_signed_vector["qr_data"],
            enc_key,
            verifier=verify_callback
        )

        expected = encrypted_signed_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.claim169.full_name == expected.get("fullName")
        assert result.verification_status == "verified"

    def test_decrypt_with_wrong_key(self, encrypted_aes256_vector):
        """Test that wrong key fails decryption."""
        wrong_key = bytes(32)  # All zeros
        with pytest.raises(claim169.Claim169Exception):
            claim169.decode_encrypted_aes(
                encrypted_aes256_vector["qr_data"], wrong_key, allow_unverified=True
            )


class TestDecodeWithVerifierHook:
    """Tests for custom verifier hook."""

    def test_custom_verifier_success(self, ed25519_signed_vector):
        """Test that custom verifier can verify signature."""
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

        public_key_bytes = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )
        public_key = Ed25519PublicKey.from_public_bytes(public_key_bytes)

        def verify_callback(algorithm, key_id, data, signature):
            # This will raise InvalidSignature if verification fails
            public_key.verify(bytes(signature), bytes(data))

        result = claim169.decode_with_verifier(
            ed25519_signed_vector["qr_data"],
            verify_callback
        )

        assert result.verification_status == "verified"

    def test_custom_verifier_failure(self, ed25519_signed_vector):
        """Test that custom verifier failure is reported."""
        def verify_callback(algorithm, key_id, data, signature):
            raise ValueError("Verification failed")

        with pytest.raises(claim169.Claim169Exception):
            claim169.decode_with_verifier(
                ed25519_signed_vector["qr_data"],
                verify_callback
            )


class TestDecodeWithDecryptorHook:
    """Tests for custom decryptor hook."""

    def test_custom_decryptor_requires_verifier_by_default(self, encrypted_aes256_vector):
        """Test that decryptor hook requires a verifier unless allow_unverified=True."""
        def decrypt_callback(algorithm, key_id, nonce, aad, ciphertext):
            return b""

        with pytest.raises(ValueError):
            claim169.decode_with_decryptor(
                encrypted_aes256_vector["qr_data"], decrypt_callback
            )

    def test_custom_decryptor_success(self, encrypted_signed_vector):
        """Test that custom decryptor can decrypt payload with nested signature."""
        from cryptography.hazmat.primitives.ciphers.aead import AESGCM
        from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

        enc_key = bytes.fromhex(
            encrypted_signed_vector["encryption_key"]["symmetric_key_hex"]
        )
        sign_key = bytes.fromhex(
            encrypted_signed_vector["signing_key"]["public_key_hex"]
        )

        aesgcm = AESGCM(enc_key)
        public_key = Ed25519PublicKey.from_public_bytes(sign_key)

        def decrypt_callback(algorithm, key_id, nonce, aad, ciphertext):
            return aesgcm.decrypt(bytes(nonce), bytes(ciphertext), bytes(aad))

        def verify_callback(algorithm, key_id, data, signature):
            public_key.verify(bytes(signature), bytes(data))

        result = claim169.decode_with_decryptor(
            encrypted_signed_vector["qr_data"],
            decrypt_callback,
            verifier=verify_callback
        )

        expected = encrypted_signed_vector["expected_claim169"]
        assert result.claim169.id == expected.get("id")
        assert result.verification_status == "verified"

    def test_custom_decryptor_failure(self, encrypted_aes256_vector):
        """Test that custom decryptor failure is reported."""
        def decrypt_callback(algorithm, key_id, nonce, aad, ciphertext):
            raise ValueError("Decryption failed")

        with pytest.raises(claim169.Claim169Exception):
            claim169.decode_with_decryptor(
                encrypted_aes256_vector["qr_data"], decrypt_callback, allow_unverified=True
            )


class TestCwtMeta:
    """Tests for CWT metadata methods."""

    def test_cwt_meta_repr(self, minimal_vector):
        """Test CwtMeta string representation."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])
        repr_str = repr(result.cwt_meta)
        assert "CwtMeta" in repr_str


class TestClaim169:
    """Tests for Claim169 methods."""

    def test_claim169_repr(self, minimal_vector):
        """Test Claim169 string representation."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])
        repr_str = repr(result.claim169)
        assert "Claim169" in repr_str

    def test_has_biometrics_true(self, with_face_vector):
        """Test has_biometrics returns True when biometrics present."""
        result = claim169.decode_unverified(with_face_vector["qr_data"])
        assert result.claim169.has_biometrics() is True

    def test_has_biometrics_false(self, minimal_vector):
        """Test has_biometrics returns False when no biometrics."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])
        assert result.claim169.has_biometrics() is False

    def test_to_dict(self, demographics_full_vector):
        """Test converting claim to dictionary."""
        result = claim169.decode_unverified(demographics_full_vector["qr_data"])
        claim_dict = result.claim169.to_dict()
        assert isinstance(claim_dict, dict)
        if result.claim169.id:
            assert claim_dict["id"] == result.claim169.id


class TestDecodeResult:
    """Tests for DecodeResult methods."""

    def test_decode_result_repr(self, minimal_vector):
        """Test DecodeResult string representation."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])
        repr_str = repr(result)
        assert "DecodeResult" in repr_str

    def test_is_verified_false_when_skipped(self, minimal_vector):
        """Test is_verified returns False when verification skipped."""
        result = claim169.decode_unverified(minimal_vector["qr_data"])
        assert result.is_verified() is False

    def test_is_verified_true_when_verified(self, ed25519_signed_vector):
        """Test is_verified returns True when verified."""
        public_key = bytes.fromhex(
            ed25519_signed_vector["signing_key"]["public_key_hex"]
        )
        result = claim169.decode_with_ed25519(
            ed25519_signed_vector["qr_data"],
            public_key
        )
        assert result.is_verified() is True



"""MOSIP Claim 169 QR Code encoder/decoder library."""

from typing import Optional

import re
from datetime import datetime

from .claim169 import (
    # Exceptions
    Claim169Exception,
    Base45DecodeError,
    DecompressError,
    CoseParseError,
    CwtParseError,
    Claim169NotFoundError,
    SignatureError,
    DecryptionError,
    EncryptionError,
    # Data classes (decode output)
    Biometric,
    CwtMeta,
    Claim169,
    DecodeResult,
    # Data classes (encode input)
    Claim169Input,
    CwtMetaInput,
    # Custom crypto provider wrappers
    PySignatureVerifier,
    PyDecryptor,
    PySigner,
    PyEncryptor,
    # Decode functions
    decode_unverified,
    decode_with_ed25519,
    decode_with_ecdsa_p256,
    decode_with_ed25519_pem,
    decode_with_ecdsa_p256_pem,
    decode_with_verifier,
    decode_encrypted_aes,
    decode_encrypted_aes256,
    decode_encrypted_aes128,
    decode_with_decryptor,
    # Encode functions
    encode_with_ed25519,
    encode_with_ecdsa_p256,
    encode_signed_encrypted,
    encode_signed_encrypted_aes128,
    encode_unsigned,
    encode_with_signer,
    encode_with_signer_and_encryptor,
    encode_with_encryptor,
    generate_nonce,
    # Utilities
    version,
)


def _validate_date_format(date: str | None, field_name: str) -> None:
    """Validate date format is ISO 8601 and represents a valid date.

    Accepts both extended format (YYYY-MM-DD) and basic format (YYYYMMDD).
    """
    if date is None:
        return

    # Accept both ISO 8601 extended (YYYY-MM-DD) and basic (YYYYMMDD) formats
    if re.match(r"^\d{4}-\d{2}-\d{2}$", date):
        date_format = "%Y-%m-%d"
    elif re.match(r"^\d{8}$", date):
        date_format = "%Y%m%d"
    else:
        raise Claim169Exception(
            f"Invalid {field_name} format: '{date}'. Expected YYYY-MM-DD or YYYYMMDD."
        )

    # Validate the date values are actually valid
    try:
        datetime.strptime(date, date_format)
    except ValueError:
        raise Claim169Exception(
            f"Invalid {field_name} value: '{date}' is not a valid calendar date."
        )

def decode(
    qr_text: str,
    skip_biometrics: bool = False,
    max_decompressed_bytes: int = 65536,
    validate_timestamps: bool = True,
    clock_skew_tolerance_seconds: int = 0,
    verify_with_ed25519: Optional[bytes] = None,
    verify_with_ecdsa_p256: Optional[bytes] = None,
    verify_with_ed25519_pem: Optional[str] = None,
    verify_with_ecdsa_p256_pem: Optional[str] = None,
    allow_unverified: bool = False,
) -> DecodeResult:
    """
    Decode a Claim 169 QR code.

    Security:
    - By default, requires signature verification via one of:
      - `verify_with_ed25519` (32 bytes)
      - `verify_with_ecdsa_p256` (33 or 65 bytes SEC1)
      - `verify_with_ed25519_pem` (PEM string)
      - `verify_with_ecdsa_p256_pem` (PEM string)
    - To explicitly decode without verification (testing only), set
      `allow_unverified=True`.
    """

    # Count how many verification methods are provided
    verifiers = [
        verify_with_ed25519 is not None,
        verify_with_ecdsa_p256 is not None,
        verify_with_ed25519_pem is not None,
        verify_with_ecdsa_p256_pem is not None,
    ]
    if sum(verifiers) > 1:
        raise ValueError("Provide only one verification key")

    result: DecodeResult

    if verify_with_ed25519 is not None:
        result = decode_with_ed25519(
            qr_text,
            verify_with_ed25519,
            skip_biometrics=skip_biometrics,
            max_decompressed_bytes=max_decompressed_bytes,
            validate_timestamps=validate_timestamps,
            clock_skew_tolerance_seconds=clock_skew_tolerance_seconds,
        )
    elif verify_with_ecdsa_p256 is not None:
        result = decode_with_ecdsa_p256(
            qr_text,
            verify_with_ecdsa_p256,
            skip_biometrics=skip_biometrics,
            max_decompressed_bytes=max_decompressed_bytes,
            validate_timestamps=validate_timestamps,
            clock_skew_tolerance_seconds=clock_skew_tolerance_seconds,
        )
    elif verify_with_ed25519_pem is not None:
        result = decode_with_ed25519_pem(
            qr_text,
            verify_with_ed25519_pem,
            skip_biometrics=skip_biometrics,
            max_decompressed_bytes=max_decompressed_bytes,
            validate_timestamps=validate_timestamps,
            clock_skew_tolerance_seconds=clock_skew_tolerance_seconds,
        )
    elif verify_with_ecdsa_p256_pem is not None:
        result = decode_with_ecdsa_p256_pem(
            qr_text,
            verify_with_ecdsa_p256_pem,
            skip_biometrics=skip_biometrics,
            max_decompressed_bytes=max_decompressed_bytes,
            validate_timestamps=validate_timestamps,
            clock_skew_tolerance_seconds=clock_skew_tolerance_seconds,
        )
    elif allow_unverified:
        result = decode_unverified(
            qr_text,
            skip_biometrics=skip_biometrics,
            max_decompressed_bytes=max_decompressed_bytes,
            validate_timestamps=validate_timestamps,
            clock_skew_tolerance_seconds=clock_skew_tolerance_seconds,
        )
    else:
        raise ValueError(
            "decode() requires a verification key (verify_with_ed25519 / verify_with_ecdsa_p256 / "
            "verify_with_ed25519_pem / verify_with_ecdsa_p256_pem) or allow_unverified=True"
        )

    # Validate date format (YYYY-MM-DD)
    _validate_date_format(result.claim169.date_of_birth, "date_of_birth")

    return result


__version__ = version()

__all__ = [
    # Exceptions
    "Claim169Exception",
    "Base45DecodeError",
    "DecompressError",
    "CoseParseError",
    "CwtParseError",
    "Claim169NotFoundError",
    "SignatureError",
    "DecryptionError",
    "EncryptionError",
    # Data classes (decode output)
    "Biometric",
    "CwtMeta",
    "Claim169",
    "DecodeResult",
    # Data classes (encode input)
    "Claim169Input",
    "CwtMetaInput",
    # Custom crypto provider wrappers
    "PySignatureVerifier",
    "PyDecryptor",
    "PySigner",
    "PyEncryptor",
    # Decode functions
    "decode_unverified",
    "decode",
    "decode_with_ed25519",
    "decode_with_ecdsa_p256",
    "decode_with_ed25519_pem",
    "decode_with_ecdsa_p256_pem",
    "decode_with_verifier",
    "decode_encrypted_aes",
    "decode_encrypted_aes256",
    "decode_encrypted_aes128",
    "decode_with_decryptor",
    # Encode functions
    "encode_with_ed25519",
    "encode_with_ecdsa_p256",
    "encode_signed_encrypted",
    "encode_signed_encrypted_aes128",
    "encode_unsigned",
    "encode_with_signer",
    "encode_with_signer_and_encryptor",
    "encode_with_encryptor",
    "generate_nonce",
    # Utilities
    "version",
    "__version__",
]
